module Hyde
  # Interchangeable pattern matching
  module PatternMatching
    def prep_path(path)
      raise Exception, "Not permitted" if @lock_methods
      @path = normalize(path) if path.kind_of? String
      @path = path if path.kind_of? Regexp
    end
    def match?(path)
      raise Exception, "Not permitted" if @lock_methods
      return true if @path == ""
      if @path.kind_of? Regexp then
        return Regexp.new("^"+@path.to_s+"$").match normalize_input(path)
      else
        split_path = path.split("/").filter { |x| x != "" }
        @path.split("/").filter { |x| x != "" }
        .each_with_index { |x,i| 
          return false if x != split_path[i]
        }
        return true
      end
    end
    def normalize_input(path)
      raise Exception, "Not permitted" if @lock_methods
      (path.split "/").filter { |x| x != "" }.join("/")
    end
    def normalize(path)
      raise Exception, "Not permitted" if @lock_methods
      # remove multiple slashes in one spot and trim edge slashes
      path = normalize_input(path)
      # globbing behaviour simulated with regexp
      if path.match /(?<!\\)[\*\?\[]/ then
        path = Regexp.new(path
          .gsub(/[\^\$\.\|\+\(\)\{\}]/,"\\\\\\0")
          .gsub(/(?<!\\)\*\*/,".*")
          .gsub(/(?<![.\\])\*/,"[^/]*")
          .gsub(/(?<!\\)\?/,".")
          .gsub(/(?<!\\)\[\!/,"[^")
        )
      end
      return path
    end
  end

  # Request control class
  class Request
    def initialize(path,request,response)
      @path = path
      @filepath = ""
      @request = request
      @response = response
    end
    attr_reader :request
    attr_reader :response
    attr_accessor :filepath
    attr_accessor :path
  end

  # Request object with safe path encapsulation
  class ProtectedRequest < Request
    def initialize(request)
      @path = request.path
      @filepath = request.filepath
      @request = request.request
      @response = request.response
    end
    undef :path=
    undef :filepath=
  end

  # Handler classes
  class Probe
    include Hyde::PatternMatching 
    def initialize (path, &block_optional) 
      prep_path path
      @block = block_optional 
    end
    def match(request)
      if match? request.path then
        p_request = Hyde::ProtectedRequest.new(request)
        @lock_methods = true
        self.instance_exec p_request, &@block if @block
        @lock_methods = false
      end
    end
  end
  class PrintProbe < Hyde::Probe
    def match(request)
      if match? request.path then
        puts "#{request.path} matched!"
      end
    end
  end 

  # Handler invocation functions
  module Handlers
    def probe(*a, **b, &block) 
      @chain.append Hyde::Probe.new *a, **b, &block
    end
    def printProbe(*a, **b, &block)
      @chain.append Hyde::PrintProbe.new *a, **b, &block
    end
  end
  
  class Pathspec
    include Hyde::PatternMatching
    include Hyde::Handlers
    def initialize (path, root_path: nil, &block)
      prep_path path
      @chain = []
      @root_override = root_path
      @remap = false
      @lock_methods = true
      self.instance_exec &block
      @lock_methods = false
    end
    def path(*a, **b, &block)
      @chain.append Hyde::Pathspec.new *a, **b, &block
    end
    def root(path)
      @lock_methods = false
      @root_override = "/"+normalize_input(path)
      @lock_methods = true
    end
    def remap(path)
      @lock_methods = false
      @root_override = "/"+normalize_input(path)
      @lock_methods = true
      @remap = true
    end
    def match(request) 
      raise Exception, "Not permitted" if @lock_methods
      if match? request.path then
        cut_path = normalize_input(request.path).sub(@path, "")
        next_path = normalize_input(request.path).match(@path)[0]
        request.path = cut_path
        if @root_override then
          request.filepath = if @remap then
                               @root_override+"/" 
                             else @root_override+next_path+"/" end
        else
          request.filepath = request.filepath+next_path+"/"
        end
        next_pathspec = @chain.find { |x| x.match? cut_path }
        next_pathspec.match request if next_pathspec
      end
    end
  end
end
